------------------------------------------------------------------------------
--                                                                          --
--                         GNAT COMPILER COMPONENTS                         --
--                                                                          --
--                              T R E E P R S                               --
--                                                                          --
--                                 S p e c                                  --
--                                                                          --
--          Copyright (C) 1992-2013, Free Software Foundation, Inc.         --
--                                                                          --
-- GNAT is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 3,  or (at your option) any later ver- --
-- sion.  GNAT is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNAT; see file COPYING3.  If not, go to --
-- http://www.gnu.org/licenses for a complete copy of the license.          --
--                                                                          --
-- GNAT was originally developed  by the GNAT team at  New York University. --
-- Extensive contributions were provided by Ada Core Technologies Inc.      --
--                                                                          --
------------------------------------------------------------------------------

--  This package contains the declaration of the string used by the Tree_Print
--  package. It must be updated whenever the arrangements of the field names
--  in package Sinfo is changed. The utility program XTREEPRS is used to
--  do this update correctly using the template treeprs.adt as input.

with Sinfo; use Sinfo;

package Treeprs is

   --------------------------------
   -- String Data for Node Print --
   --------------------------------

   --  String data for print out. The Pchars array is a long string with the
   --  the entry for each node type consisting of a single blank, followed by
   --  a series of entries, one for each Op or Flag field used for the node.
   --  Each entry has a single character which identifies the field, followed
   --  by the synonym name. The starting location for a given node type is
   --  found from the corresponding entry in the Pchars_Pos_Array.

   --  The following characters identify the field. These are characters which
   --  could never occur in a field name, so they also mark the end of the
   --  previous name.

   --  Note the following definitions do not include Flag0. This will have to
   --  be addressed if we ever need to use Flag0 (it's not currently used).

   subtype Fchar is Character range '#' .. '9';

   F_Field1     : constant Fchar := '#'; -- Character'Val (16#23#)
   F_Field2     : constant Fchar := '$'; -- Character'Val (16#24#)
   F_Field3     : constant Fchar := '%'; -- Character'Val (16#25#)
   F_Field4     : constant Fchar := '&'; -- Character'Val (16#26#)
   F_Field5     : constant Fchar := '''; -- Character'Val (16#27#)
   F_Flag1      : constant Fchar := '('; -- Character'Val (16#28#)
   F_Flag2      : constant Fchar := ')'; -- Character'Val (16#29#)
   F_Flag3      : constant Fchar := '*'; -- Character'Val (16#2A#)
   F_Flag4      : constant Fchar := '+'; -- Character'Val (16#2B#)
   F_Flag5      : constant Fchar := ','; -- Character'Val (16#2C#)
   F_Flag6      : constant Fchar := '-'; -- Character'Val (16#2D#)
   F_Flag7      : constant Fchar := '.'; -- Character'Val (16#2E#)
   F_Flag8      : constant Fchar := '/'; -- Character'Val (16#2F#)
   F_Flag9      : constant Fchar := '0'; -- Character'Val (16#30#)
   F_Flag10     : constant Fchar := '1'; -- Character'Val (16#31#)
   F_Flag11     : constant Fchar := '2'; -- Character'Val (16#32#)
   F_Flag12     : constant Fchar := '3'; -- Character'Val (16#33#)
   F_Flag13     : constant Fchar := '4'; -- Character'Val (16#34#)
   F_Flag14     : constant Fchar := '5'; -- Character'Val (16#35#)
   F_Flag15     : constant Fchar := '6'; -- Character'Val (16#36#)
   F_Flag16     : constant Fchar := '7'; -- Character'Val (16#37#)
   F_Flag17     : constant Fchar := '8'; -- Character'Val (16#38#)
   F_Flag18     : constant Fchar := '9'; -- Character'Val (16#39#)

   --  Note this table does not include entity field and flags whose access
   --  functions are in Einfo (these are handled by the Print_Entity_Info
   --  procedure in Treepr, which uses the routines in Einfo to get the proper
   --  symbolic information). In addition, the following fields are handled by
   --  Treepr, and do not appear in the Pchars array:

   --    Analyzed
   --    Cannot_Be_Constant
   --    Chars
   --    Comes_From_Source
   --    Error_Posted
   --    Etype
   --    Is_Controlling_Actual
   --    Is_Overloaded
   --    Is_Static_Expression
   --    Left_Opnd
   --    Must_Check_Expr
   --    Must_Not_Freeze
   --    No_Overflow_Expr
   --    Paren_Count
   --    Raises_Constraint_Error
   --    Right_Opnd

   Pchars : constant String :=
      --  Unused_At_Start
      "" &
      --  At_Clause
      "#Identifier%Expression" &
      --  Component_Clause
      "#Component_Name$Position%First_Bit&Last_Bit" &
      --  Enumeration_Representation_Clause
      "#Identifier%Array_Aggregate'Next_Rep_Item" &
      --  Mod_Clause
      "%Expression&Pragmas_Before" &
      --  Record_Representation_Clause
      "#Identifier$Mod_Clause%Component_Clauses'Next_Rep_Item" &
      --  Attribute_Definition_Clause
      "$Name%Expression&Entity'Next_Rep_Item+From_At_Mod2Check_Address_Align" &
         "ment4From_Aspect_Specification5Is_Delayed_Aspect9Address_Warning_P" &
         "osted" &
      --  Empty
      "" &
      --  Pragma_Argument_Association
      "$Expression_Copy%Expression" &
      --  Error
      "" &
      --  Defining_Character_Literal
      "$Next_Entity%Scope" &
      --  Defining_Identifier
      "$Next_Entity%Scope" &
      --  Defining_Operator_Symbol
      "$Next_Entity%Scope" &
      --  Expanded_Name
      "%Prefix$Selector_Name&Entity&Associated_Node(Is_Elaboration_Checks_OK" &
         "_Node)Is_SPARK_Mode_On_Node*Is_Elaboration_Warnings_OK_Node2Has_Pr" &
         "ivate_View4Redundant_Use5Atomic_Sync_Required" &
      --  Identifier
      "&Entity&Associated_Node$Original_Discriminant(Is_Elaboration_Checks_O" &
         "K_Node)Is_SPARK_Mode_On_Node*Is_Elaboration_Warnings_OK_Node2Has_P" &
         "rivate_View4Redundant_Use5Atomic_Sync_Required" &
      --  Operator_Symbol
      "%Strval&Entity&Associated_Node2Has_Private_View" &
      --  Character_Literal
      "$Char_Literal_Value&Entity&Associated_Node2Has_Private_View" &
      --  Op_Add
      "" &
      --  Op_Concat
      "4Is_Component_Left_Opnd5Is_Component_Right_Opnd" &
      --  Op_Expon
      "4Is_Power_Of_2_For_Shift" &
      --  Op_Subtract
      "" &
      --  Op_Divide
      "5Treat_Fixed_As_Integer4Do_Division_Check9Rounded_Result" &
      --  Op_Mod
      "5Treat_Fixed_As_Integer4Do_Division_Check" &
      --  Op_Multiply
      "5Treat_Fixed_As_Integer9Rounded_Result" &
      --  Op_Rem
      "5Treat_Fixed_As_Integer4Do_Division_Check" &
      --  Op_And
      "+Do_Length_Check" &
      --  Op_Eq
      "" &
      --  Op_Ge
      "" &
      --  Op_Gt
      "" &
      --  Op_Le
      "" &
      --  Op_Lt
      "" &
      --  Op_Ne
      "" &
      --  Op_Or
      "+Do_Length_Check" &
      --  Op_Xor
      "+Do_Length_Check" &
      --  Op_Rotate_Left
      "+Shift_Count_OK" &
      --  Op_Rotate_Right
      "+Shift_Count_OK" &
      --  Op_Shift_Left
      "+Shift_Count_OK" &
      --  Op_Shift_Right
      "+Shift_Count_OK" &
      --  Op_Shift_Right_Arithmetic
      "+Shift_Count_OK" &
      --  Op_Abs
      "" &
      --  Op_Minus
      "" &
      --  Op_Not
      "" &
      --  Op_Plus
      "" &
      --  Attribute_Reference
      "%Prefix$Attribute_Name#Expressions&Entity&Associated_Node(Is_Elaborat" &
         "ion_Checks_OK_Node)Is_SPARK_Mode_On_Node*Is_Elaboration_Warnings_O" &
         "K_Node2Header_Size_Added4Redundant_Use5Must_Be_Byte_Aligned" &
      --  In
      "&Alternatives8No_Minimize_Eliminate" &
      --  Not_In
      "&Alternatives8No_Minimize_Eliminate" &
      --  And_Then
      "#Actions" &
      --  Or_Else
      "#Actions" &
      --  Function_Call
      "$Name%Parameter_Associations&First_Named_Actual#Controlling_Argument(" &
         "Is_Elaboration_Checks_OK_Node)Is_SPARK_Mode_On_Node*Is_Elaboration" &
         "_Warnings_OK_Node+No_Elaboration_Check2Is_Expanded_Build_In_Place_" &
         "Call4Do_Tag_Check8No_Side_Effect_Removal9Is_Known_Guaranteed_ABE" &
      --  Procedure_Call_Statement
      "$Name%Parameter_Associations&First_Named_Actual#Controlling_Argument(" &
         "Is_Elaboration_Checks_OK_Node)Is_SPARK_Mode_On_Node*Is_Elaboration" &
         "_Warnings_OK_Node+No_Elaboration_Check4Do_Tag_Check9Is_Known_Guara" &
         "nteed_ABE" &
      --  Raise_Constraint_Error
      "#Condition%Reason" &
      --  Raise_Program_Error
      "#Condition%Reason" &
      --  Raise_Storage_Error
      "#Condition%Reason" &
      --  Integer_Literal
      "$Original_Entity%Intval4Print_In_Hex" &
      --  Real_Literal
      "$Original_Entity%Realval&Corresponding_Integer_Value2Is_Machine_Numbe" &
         "r" &
      --  String_Literal
      "%Strval2Has_Wide_Character4Has_Wide_Wide_Character+Is_Folded_In_Parse" &
         "r" &
      --  Explicit_Dereference
      "%Prefix&Actual_Designated_Subtype4Has_Dereference_Action5Atomic_Sync_" &
         "Required" &
      --  Expression_With_Actions
      "#Actions%Expression" &
      --  If_Expression
      "#Expressions$Then_Actions%Else_Actions4Is_Elsif8Do_Overflow_Check" &
      --  Indexed_Component
      "%Prefix#Expressions&Generalized_Indexing5Atomic_Sync_Required" &
      --  Null
      "" &
      --  Qualified_Expression
      "&Subtype_Mark%Expression+Is_Qualified_Universal_Literal" &
      --  Quantified_Expression
      "$Iterator_Specification&Loop_Parameter_Specification#Condition6All_Pr" &
         "esent" &
      --  Aggregate
      "#Expressions$Component_Associations8Null_Record_Present%Aggregate_Bou" &
         "nds&Associated_Node9Compile_Time_Known_Aggregate2Expansion_Delayed" &
         "4Has_Self_Reference5Is_Homogeneous_Aggregate" &
      --  Allocator
      "%Expression&Subpool_Handle_Name#Storage_Pool$Procedure_To_Call(Alloc_" &
         "For_BIP_Return2Null_Exclusion_Present4No_Initialization5Is_Static_" &
         "Coextension8Do_Storage_Check9Is_Dynamic_Coextension" &
      --  Case_Expression
      "%Expression&Alternatives8Do_Overflow_Check" &
      --  Delta_Aggregate
      "%Expression$Component_Associations" &
      --  Extension_Aggregate
      "%Ancestor_Part&Associated_Node#Expressions$Component_Associations8Nul" &
         "l_Record_Present2Expansion_Delayed4Has_Self_Reference" &
      --  Raise_Expression
      "$Name%Expression4Convert_To_Return_False" &
      --  Range
      "#Low_Bound$High_Bound2Includes_Infinities" &
      --  Reference
      "%Prefix" &
      --  Selected_Component
      "%Prefix$Selector_Name&Associated_Node*Do_Discriminant_Check2Is_In_Dis" &
         "criminant_Check5Atomic_Sync_Required8Is_Prefixed_Call" &
      --  Slice
      "%Prefix&Discrete_Range" &
      --  Target_Name
      "" &
      --  Type_Conversion
      "&Subtype_Mark%Expression*Do_Discriminant_Check+Do_Length_Check2Float_" &
         "Truncate4Do_Tag_Check5Conversion_OK8Do_Overflow_Check9Rounded_Resu" &
         "lt" &
      --  Unchecked_Expression
      "%Expression" &
      --  Unchecked_Type_Conversion
      "&Subtype_Mark%Expression2Kill_Range_Check8No_Truncation" &
      --  Subtype_Indication
      "&Subtype_Mark%Constraint/Must_Not_Freeze" &
      --  Component_Declaration
      "#Defining_Identifier&Component_Definition%Expression,More_Ids-Prev_Id" &
         "s" &
      --  Entry_Declaration
      "#Defining_Identifier&Discrete_Subtype_Definition%Parameter_Specificat" &
         "ions'Corresponding_Body5Must_Override6Must_Not_Override" &
      --  Expression_Function
      "#Specification%Expression'Corresponding_Spec" &
      --  Formal_Object_Declaration
      "#Defining_Identifier6In_Present8Out_Present2Null_Exclusion_Present&Su" &
         "btype_Mark%Access_Definition'Default_Expression,More_Ids-Prev_Ids" &
      --  Formal_Type_Declaration
      "#Defining_Identifier%Formal_Type_Definition&Discriminant_Specificatio" &
         "ns4Unknown_Discriminants_Present" &
      --  Full_Type_Declaration
      "#Defining_Identifier$Incomplete_View&Discriminant_Specifications%Type" &
         "_Definition2Discr_Check_Funcs_Built" &
      --  Incomplete_Type_Declaration
      "#Defining_Identifier&Discriminant_Specifications'Premature_Use4Unknow" &
         "n_Discriminants_Present6Tagged_Present" &
      --  Iterator_Specification
      "#Defining_Identifier$Name6Reverse_Present7Of_Present'Subtype_Indicati" &
         "on" &
      --  Loop_Parameter_Specification
      "#Defining_Identifier6Reverse_Present&Discrete_Subtype_Definition" &
      --  Object_Declaration
      "#Defining_Identifier+Aliased_Present8Constant_Present2Null_Exclusion_" &
         "Present&Object_Definition%Expression$Handler_List_Entry'Correspond" &
         "ing_Generic_Association,More_Ids-Prev_Ids4No_Initialization6Assign" &
         "ment_OK/Exception_Junk7Is_Subprogram_Descriptor5Has_Init_Expressio" &
         "n9Suppress_Assignment_Checks" &
      --  Protected_Type_Declaration
      "#Defining_Identifier&Discriminant_Specifications$Interface_List%Prote" &
         "cted_Definition'Corresponding_Body" &
      --  Private_Extension_Declaration
      "#Defining_Identifier%Uninitialized_Variable&Discriminant_Specificatio" &
         "ns4Unknown_Discriminants_Present+Abstract_Present8Limited_Present." &
         "Synchronized_Present'Subtype_Indication$Interface_List" &
      --  Private_Type_Declaration
      "#Defining_Identifier&Discriminant_Specifications4Unknown_Discriminant" &
         "s_Present+Abstract_Present6Tagged_Present8Limited_Present" &
      --  Subtype_Declaration
      "#Defining_Identifier2Null_Exclusion_Present'Subtype_Indication&Generi" &
         "c_Parent_Type/Exception_Junk3Has_Dynamic_Range_Check" &
      --  Function_Specification
      "#Defining_Unit_Name%Parameter_Specifications2Null_Exclusion_Present&R" &
         "esult_Definition'Generic_Parent5Must_Override6Must_Not_Override" &
      --  Procedure_Specification
      "#Defining_Unit_Name$Null_Statement%Parameter_Specifications'Generic_P" &
         "arent4Null_Present5Must_Override6Must_Not_Override" &
      --  Access_Function_Definition
      "2Null_Exclusion_Present5Null_Exclusion_In_Return_Present-Protected_Pr" &
         "esent%Parameter_Specifications&Result_Definition" &
      --  Access_Procedure_Definition
      "2Null_Exclusion_Present-Protected_Present%Parameter_Specifications" &
      --  Task_Type_Declaration
      "#Defining_Identifier&Discriminant_Specifications$Interface_List%Task_" &
         "Definition'Corresponding_Body" &
      --  Package_Body_Stub
      "#Defining_Identifier$Corresponding_Spec_Of_Stub&Library_Unit'Correspo" &
         "nding_Body" &
      --  Protected_Body_Stub
      "#Defining_Identifier$Corresponding_Spec_Of_Stub&Library_Unit'Correspo" &
         "nding_Body" &
      --  Subprogram_Body_Stub
      "#Specification$Corresponding_Spec_Of_Stub&Library_Unit'Corresponding_" &
         "Body" &
      --  Task_Body_Stub
      "#Defining_Identifier$Corresponding_Spec_Of_Stub&Library_Unit'Correspo" &
         "nding_Body" &
      --  Function_Instantiation
      "#Defining_Unit_Name$Name%Generic_Associations&Parent_Spec'Instance_Sp" &
         "ec(Is_Elaboration_Checks_OK_Node)Is_SPARK_Mode_On_Node*Is_Elaborat" &
         "ion_Warnings_OK_Node,Is_Declaration_Level_Node5Must_Override6Must_" &
         "Not_Override9Is_Known_Guaranteed_ABE" &
      --  Procedure_Instantiation
      "#Defining_Unit_Name$Name&Parent_Spec%Generic_Associations'Instance_Sp" &
         "ec(Is_Elaboration_Checks_OK_Node)Is_SPARK_Mode_On_Node*Is_Elaborat" &
         "ion_Warnings_OK_Node,Is_Declaration_Level_Node5Must_Override6Must_" &
         "Not_Override9Is_Known_Guaranteed_ABE" &
      --  Package_Instantiation
      "#Defining_Unit_Name$Name%Generic_Associations&Parent_Spec'Instance_Sp" &
         "ec(Is_Elaboration_Checks_OK_Node)Is_SPARK_Mode_On_Node*Is_Elaborat" &
         "ion_Warnings_OK_Node,Is_Declaration_Level_Node9Is_Known_Guaranteed" &
         "_ABE" &
      --  Package_Body
      "#Defining_Unit_Name$Declarations&Handled_Statement_Sequence'Correspon" &
         "ding_Spec4Was_Originally_Stub" &
      --  Subprogram_Body
      "#Specification$Declarations&Handled_Statement_Sequence%Activation_Cha" &
         "in_Entity'Corresponding_Spec+Acts_As_Spec6Bad_Is_Detected8Do_Stora" &
         "ge_Check0Has_Relative_Deadline_Pragma/Is_Entry_Barrier_Function.Is" &
         "_Protected_Subprogram_Body(Is_Task_Body_Procedure,Is_Task_Master)W" &
         "as_Attribute_Reference9Was_Expression_Function4Was_Originally_Stub" &
      --  Protected_Body
      "#Defining_Identifier$Declarations&End_Label'Corresponding_Spec4Was_Or" &
         "iginally_Stub" &
      --  Task_Body
      "#Defining_Identifier$Declarations&Handled_Statement_Sequence,Is_Task_" &
         "Master%Activation_Chain_Entity'Corresponding_Spec4Was_Originally_S" &
         "tub" &
      --  Implicit_Label_Declaration
      "#Defining_Identifier$Label_Construct" &
      --  Package_Declaration
      "#Specification'Corresponding_Body&Parent_Spec%Activation_Chain_Entity" &
      --  Single_Task_Declaration
      "#Defining_Identifier$Interface_List%Task_Definition" &
      --  Subprogram_Declaration
      "#Specification%Body_To_Inline'Corresponding_Body&Parent_Spec/Is_Entry" &
         "_Barrier_Function(Is_Task_Body_Procedure" &
      --  Use_Package_Clause
      "#Prev_Use_Clause$Name%Next_Use_Clause&Associated_Node'Hidden_By_Use_C" &
         "lause(Is_Effective_Use_Clause,More_Ids-Prev_Ids" &
      --  Generic_Package_Declaration
      "#Specification'Corresponding_Body$Generic_Formal_Declarations&Parent_" &
         "Spec%Activation_Chain_Entity" &
      --  Generic_Subprogram_Declaration
      "#Specification'Corresponding_Body$Generic_Formal_Declarations&Parent_" &
         "Spec" &
      --  Constrained_Array_Definition
      "$Discrete_Subtype_Definitions&Component_Definition" &
      --  Unconstrained_Array_Definition
      "$Subtype_Marks&Component_Definition" &
      --  Exception_Renaming_Declaration
      "#Defining_Identifier$Name" &
      --  Object_Renaming_Declaration
      "#Defining_Identifier2Null_Exclusion_Present&Subtype_Mark%Access_Defin" &
         "ition$Name'Corresponding_Generic_Association" &
      --  Package_Renaming_Declaration
      "#Defining_Unit_Name$Name&Parent_Spec" &
      --  Subprogram_Renaming_Declaration
      "#Specification$Name&Parent_Spec'Corresponding_Spec%Corresponding_Form" &
         "al_Spec-From_Default" &
      --  Generic_Function_Renaming_Declaration
      "#Defining_Unit_Name$Name&Parent_Spec" &
      --  Generic_Package_Renaming_Declaration
      "#Defining_Unit_Name$Name&Parent_Spec" &
      --  Generic_Procedure_Renaming_Declaration
      "#Defining_Unit_Name$Name&Parent_Spec" &
      --  Abort_Statement
      "$Names" &
      --  Accept_Statement
      "#Entry_Direct_Name'Entry_Index%Parameter_Specifications&Handled_State" &
         "ment_Sequence$Declarations" &
      --  Assignment_Statement
      "$Name%Expression(Is_Elaboration_Checks_OK_Node)Is_SPARK_Mode_On_Node*" &
         "Do_Discriminant_Check+Do_Length_Check,Forwards_OK-Backwards_OK.No_" &
         "Ctrl_Actions/Has_Target_Names0Is_Elaboration_Code4Do_Tag_Check5Com" &
         "ponentwise_Assignment9Suppress_Assignment_Checks" &
      --  Asynchronous_Select
      "#Triggering_Alternative$Abortable_Part" &
      --  Block_Statement
      "#Identifier$Declarations&Handled_Statement_Sequence%Activation_Chain_" &
         "Entity'Cleanup_Actions6Has_Created_Identifier.Is_Asynchronous_Call" &
         "_Block-Is_Task_Allocation_Block/Exception_Junk+Is_Abort_Block0Is_F" &
         "inalization_Wrapper(Is_Initialization_Block,Is_Task_Master" &
      --  Case_Statement
      "%Expression&Alternatives'End_Span(From_Conditional_Expression" &
      --  Code_Statement
      "%Expression" &
      --  Compound_Statement
      "#Actions" &
      --  Conditional_Entry_Call
      "#Entry_Call_Alternative&Else_Statements" &
      --  Delay_Relative_Statement
      "%Expression" &
      --  Delay_Until_Statement
      "%Expression" &
      --  Entry_Call_Statement
      "$Name%Parameter_Associations&First_Named_Actual(Is_Elaboration_Checks" &
         "_OK_Node)Is_SPARK_Mode_On_Node*Is_Elaboration_Warnings_OK_Node" &
      --  Free_Statement
      "%Expression#Storage_Pool$Procedure_To_Call&Actual_Designated_Subtype" &
      --  Goto_Statement
      "$Name/Exception_Junk" &
      --  Loop_Statement
      "#Identifier$Iteration_Scheme%Statements&End_Label4Is_OpenAcc_Environm" &
         "ent5Is_OpenAcc_Loop6Has_Created_Identifier7Is_Null_Loop8Suppress_L" &
         "oop_Warnings" &
      --  Null_Statement
      "" &
      --  Raise_Statement
      "$Name%Expression+From_At_End" &
      --  Requeue_Statement
      "$Name6Abort_Present(Is_Elaboration_Checks_OK_Node)Is_SPARK_Mode_On_No" &
         "de*Is_Elaboration_Warnings_OK_Node" &
      --  Simple_Return_Statement
      "'Return_Statement_Entity%Expression#Storage_Pool$Procedure_To_Call4Do" &
         "_Tag_Check,By_Ref9Comes_From_Extended_Return_Statement" &
      --  Extended_Return_Statement
      "'Return_Statement_Entity%Return_Object_Declarations&Handled_Statement" &
         "_Sequence#Storage_Pool$Procedure_To_Call4Do_Tag_Check,By_Ref" &
      --  Selective_Accept
      "#Select_Alternatives&Else_Statements" &
      --  Timed_Entry_Call
      "#Entry_Call_Alternative&Delay_Alternative" &
      --  Exit_Statement
      "$Name#Condition%Next_Exit_Statement" &
      --  If_Statement
      "#Condition$Then_Statements%Elsif_Parts&Else_Statements'End_Span(From_" &
         "Conditional_Expression" &
      --  Accept_Alternative
      "$Accept_Statement#Condition%Statements&Pragmas_Before'Accept_Handler_" &
         "Records" &
      --  Delay_Alternative
      "$Delay_Statement#Condition%Statements&Pragmas_Before" &
      --  Elsif_Part
      "#Condition$Then_Statements%Condition_Actions" &
      --  Entry_Body_Formal_Part
      "&Entry_Index_Specification%Parameter_Specifications#Condition" &
      --  Iteration_Scheme
      "#Condition%Condition_Actions$Iterator_Specification&Loop_Parameter_Sp" &
         "ecification" &
      --  Terminate_Alternative
      "#Condition&Pragmas_Before'Pragmas_After" &
      --  Formal_Abstract_Subprogram_Declaration
      "#Specification$Default_Name6Box_Present" &
      --  Formal_Concrete_Subprogram_Declaration
      "#Specification$Default_Name6Box_Present" &
      --  Push_Constraint_Error_Label
      "'Exception_Label" &
      --  Push_Program_Error_Label
      "'Exception_Label" &
      --  Push_Storage_Error_Label
      "'Exception_Label" &
      --  Pop_Constraint_Error_Label
      "" &
      --  Pop_Program_Error_Label
      "" &
      --  Pop_Storage_Error_Label
      "" &
      --  SCIL_Dispatch_Table_Tag_Init
      "&SCIL_Entity" &
      --  SCIL_Dispatching_Call
      "$SCIL_Target_Prim&SCIL_Entity'SCIL_Controlling_Tag" &
      --  SCIL_Membership_Test
      "'SCIL_Tag_Value&SCIL_Entity" &
      --  Abortable_Part
      "%Statements" &
      --  Abstract_Subprogram_Declaration
      "#Specification" &
      --  Access_Definition
      "2Null_Exclusion_Present6All_Present8Constant_Present&Subtype_Mark%Acc" &
         "ess_To_Subprogram_Definition" &
      --  Access_To_Object_Definition
      "6All_Present2Null_Exclusion_Present7Null_Excluding_Subtype'Subtype_In" &
         "dication8Constant_Present" &
      --  Aspect_Specification
      "#Identifier$Aspect_Rep_Item%Expression&Entity'Next_Rep_Item-Class_Pre" &
         "sent0Is_Ignored2Is_Checked5Is_Delayed_Aspect6Is_Disabled7Is_Boolea" &
         "n_Aspect8Split_PPC9Aspect_On_Partial_View" &
      --  Call_Marker
      "#Target(Is_Elaboration_Checks_OK_Node)Is_SPARK_Mode_On_Node*Is_Elabor" &
         "ation_Warnings_OK_Node+Is_Source_Call,Is_Declaration_Level_Node-Is" &
         "_Dispatching_Call9Is_Known_Guaranteed_ABE" &
      --  Case_Expression_Alternative
      "#Actions&Discrete_Choices%Expression6Has_SP_Choice" &
      --  Case_Statement_Alternative
      "&Discrete_Choices%Statements6Has_SP_Choice" &
      --  Compilation_Unit
      "#Context_Items6Private_Present$Unit'Aux_Decls_Node%First_Inlined_Subp" &
         "rogram&Library_Unit(Save_Invocation_Graph_Of_Body+Acts_As_Spec4Bod" &
         "y_Required5Has_Pragma_Suppress_All7Context_Pending8Has_No_Elaborat" &
         "ion_Code" &
      --  Compilation_Unit_Aux
      "$Declarations#Actions'Pragmas_After&Config_Pragmas%Default_Storage_Po" &
         "ol" &
      --  Component_Association
      "#Choices$Loop_Actions%Expression6Box_Present4Inherited_Discriminant" &
      --  Component_Definition
      "+Aliased_Present2Null_Exclusion_Present'Subtype_Indication%Access_Def" &
         "inition" &
      --  Component_List
      "%Component_Items&Variant_Part4Null_Present" &
      --  Contract
      "#Pre_Post_Conditions$Contract_Test_Cases%Classifications(Is_Expanded_" &
         "Contract" &
      --  Derived_Type_Definition
      "+Abstract_Present2Null_Exclusion_Present'Subtype_Indication%Record_Ex" &
         "tension_Part8Limited_Present,Task_Present-Protected_Present.Synchr" &
         "onized_Present$Interface_List7Interface_Present" &
      --  Decimal_Fixed_Point_Definition
      "%Delta_Expression$Digits_Expression&Real_Range_Specification" &
      --  Defining_Program_Unit_Name
      "$Name#Defining_Identifier" &
      --  Delta_Constraint
      "%Delta_Expression&Range_Constraint" &
      --  Designator
      "$Name#Identifier" &
      --  Digits_Constraint
      "$Digits_Expression&Range_Constraint" &
      --  Discriminant_Association
      "#Selector_Names%Expression" &
      --  Discriminant_Specification
      "#Defining_Identifier2Null_Exclusion_Present'Discriminant_Type%Express" &
         "ion,More_Ids-Prev_Ids" &
      --  Enumeration_Type_Definition
      "#Literals&End_Label" &
      --  Entry_Body
      "#Defining_Identifier'Entry_Body_Formal_Part$Declarations&Handled_Stat" &
         "ement_Sequence%Activation_Chain_Entity" &
      --  Entry_Call_Alternative
      "#Entry_Call_Statement%Statements&Pragmas_Before" &
      --  Entry_Index_Specification
      "#Defining_Identifier&Discrete_Subtype_Definition" &
      --  Exception_Declaration
      "#Defining_Identifier%Expression$Renaming_Exception,More_Ids-Prev_Ids" &
      --  Exception_Handler
      "$Choice_Parameter&Exception_Choices%Statements'Exception_Label#Local_" &
         "Raise_Statements.Local_Raise_Not_OK/Has_Local_Raise" &
      --  Floating_Point_Definition
      "$Digits_Expression&Real_Range_Specification" &
      --  Formal_Decimal_Fixed_Point_Definition
      "" &
      --  Formal_Derived_Type_Definition
      "&Subtype_Mark6Private_Present+Abstract_Present8Limited_Present.Synchr" &
         "onized_Present$Interface_List" &
      --  Formal_Discrete_Type_Definition
      "" &
      --  Formal_Floating_Point_Definition
      "" &
      --  Formal_Modular_Type_Definition
      "" &
      --  Formal_Ordinary_Fixed_Point_Definition
      "" &
      --  Formal_Package_Declaration
      "#Defining_Identifier$Name%Generic_Associations6Box_Present'Instance_S" &
         "pec9Is_Known_Guaranteed_ABE" &
      --  Formal_Private_Type_Definition
      "%Uninitialized_Variable+Abstract_Present6Tagged_Present8Limited_Prese" &
         "nt" &
      --  Formal_Incomplete_Type_Definition
      "6Tagged_Present" &
      --  Formal_Signed_Integer_Type_Definition
      "" &
      --  Freeze_Entity
      "&Entity$Access_Types_To_Process%TSS_Elist#Actions'First_Subtype_Link" &
      --  Freeze_Generic_Entity
      "&Entity" &
      --  Generic_Association
      "$Selector_Name#Explicit_Generic_Actual_Parameter6Box_Present" &
      --  Handled_Sequence_Of_Statements
      "%Statements&End_Label'Exception_Handlers#At_End_Proc$First_Real_State" &
         "ment" &
      --  Index_Or_Discriminant_Constraint
      "#Constraints" &
      --  Iterated_Component_Association
      "#Defining_Identifier$Loop_Actions%Expression&Discrete_Choices6Box_Pre" &
         "sent" &
      --  Itype_Reference
      "#Itype" &
      --  Label
      "#Identifier/Exception_Junk" &
      --  Modular_Type_Definition
      "%Expression" &
      --  Number_Declaration
      "#Defining_Identifier%Expression,More_Ids-Prev_Ids" &
      --  Ordinary_Fixed_Point_Definition
      "%Delta_Expression&Real_Range_Specification" &
      --  Others_Choice
      "#Others_Discrete_Choices2All_Others" &
      --  Package_Specification
      "#Defining_Unit_Name$Visible_Declarations%Private_Declarations&End_Lab" &
         "el'Generic_Parent9Limited_View_Installed" &
      --  Parameter_Association
      "$Selector_Name%Explicit_Actual_Parameter&Next_Named_Actual4Is_Accessi" &
         "bility_Actual" &
      --  Parameter_Specification
      "#Defining_Identifier+Aliased_Present6In_Present8Out_Present2Null_Excl" &
         "usion_Present$Parameter_Type%Expression4Do_Accessibility_Check,Mor" &
         "e_Ids-Prev_Ids'Default_Expression" &
      --  Pragma
      "#Next_Pragma$Pragma_Argument_Associations%Corresponding_Aspect&Pragma" &
         "_Identifier'Next_Rep_Item)Is_Generic_Contract_Pragma*Is_Checked_Gh" &
         "ost_Pragma+Is_Inherited_Pragma,Is_Analyzed_Pragma-Class_Present.Un" &
         "eval_Old_Accept/Is_Ignored_Ghost_Pragma0Is_Ignored2Is_Checked4From" &
         "_Aspect_Specification5Is_Delayed_Aspect6Is_Disabled7Import_Interfa" &
         "ce_Present8Split_PPC9Uneval_Old_Warn" &
      --  Protected_Definition
      "$Visible_Declarations%Private_Declarations&End_Label" &
      --  Range_Constraint
      "&Range_Expression" &
      --  Real_Range_Specification
      "#Low_Bound$High_Bound" &
      --  Record_Definition
      "&End_Label+Abstract_Present6Tagged_Present8Limited_Present#Component_" &
         "List4Null_Present,Task_Present-Protected_Present.Synchronized_Pres" &
         "ent7Interface_Present$Interface_List" &
      --  Signed_Integer_Type_Definition
      "#Low_Bound$High_Bound" &
      --  Single_Protected_Declaration
      "#Defining_Identifier$Interface_List%Protected_Definition" &
      --  Subunit
      "$Name#Proper_Body%Corresponding_Stub" &
      --  Task_Definition
      "$Visible_Declarations%Private_Declarations&End_Label,Has_Storage_Size" &
         "_Pragma0Has_Relative_Deadline_Pragma" &
      --  Triggering_Alternative
      "#Triggering_Statement%Statements&Pragmas_Before" &
      --  Use_Type_Clause
      "#Prev_Use_Clause$Used_Operations%Next_Use_Clause&Subtype_Mark'Hidden_" &
         "By_Use_Clause(Is_Effective_Use_Clause,More_Ids-Prev_Ids6All_Presen" &
         "t" &
      --  Validate_Unchecked_Conversion
      "#Source_Type$Target_Type" &
      --  Variable_Reference_Marker
      "#Target(Is_Elaboration_Checks_OK_Node)Is_SPARK_Mode_On_Node*Is_Elabor" &
         "ation_Warnings_OK_Node+Is_Read,Is_Write" &
      --  Variant
      "&Discrete_Choices#Component_List$Enclosing_Variant%Present_Expr'Dchec" &
         "k_Function6Has_SP_Choice" &
      --  Variant_Part
      "$Name#Variants" &
      --  With_Clause
      "$Name6Private_Present8Limited_Present%Next_Implicit_With&Library_Unit" &
         "'Corresponding_Spec,First_Name-Last_Name4Context_Installed+Elabora" &
         "te_Present5Elaborate_All_Present0Elaborate_All_Desirable2Elaborate" &
         "_Desirable7Implicit_With9Limited_View_Installed(Parent_With.Unrefe" &
         "renced_In_Spec/No_Entities_Ref_In_Spec" &
      --  Unused_At_End
      "";

   type Pchar_Pos_Array is array (Node_Kind) of Positive;
   Pchar_Pos : constant Pchar_Pos_Array := Pchar_Pos_Array'(
      N_Unused_At_Start                        => 1,
      N_At_Clause                              => 1,
      N_Component_Clause                       => 23,
      N_Enumeration_Representation_Clause      => 66,
      N_Mod_Clause                             => 107,
      N_Record_Representation_Clause           => 133,
      N_Attribute_Definition_Clause            => 187,
      N_Empty                                  => 327,
      N_Pragma_Argument_Association            => 327,
      N_Error                                  => 354,
      N_Defining_Character_Literal             => 354,
      N_Defining_Identifier                    => 372,
      N_Defining_Operator_Symbol               => 390,
      N_Expanded_Name                          => 408,
      N_Identifier                             => 588,
      N_Operator_Symbol                        => 769,
      N_Character_Literal                      => 816,
      N_Op_Add                                 => 875,
      N_Op_Concat                              => 875,
      N_Op_Expon                               => 922,
      N_Op_Subtract                            => 946,
      N_Op_Divide                              => 946,
      N_Op_Mod                                 => 1002,
      N_Op_Multiply                            => 1043,
      N_Op_Rem                                 => 1081,
      N_Op_And                                 => 1122,
      N_Op_Eq                                  => 1138,
      N_Op_Ge                                  => 1138,
      N_Op_Gt                                  => 1138,
      N_Op_Le                                  => 1138,
      N_Op_Lt                                  => 1138,
      N_Op_Ne                                  => 1138,
      N_Op_Or                                  => 1138,
      N_Op_Xor                                 => 1154,
      N_Op_Rotate_Left                         => 1170,
      N_Op_Rotate_Right                        => 1185,
      N_Op_Shift_Left                          => 1200,
      N_Op_Shift_Right                         => 1215,
      N_Op_Shift_Right_Arithmetic              => 1230,
      N_Op_Abs                                 => 1245,
      N_Op_Minus                               => 1245,
      N_Op_Not                                 => 1245,
      N_Op_Plus                                => 1245,
      N_Attribute_Reference                    => 1245,
      N_In                                     => 1439,
      N_Not_In                                 => 1474,
      N_And_Then                               => 1509,
      N_Or_Else                                => 1517,
      N_Function_Call                          => 1525,
      N_Procedure_Call_Statement               => 1790,
      N_Raise_Constraint_Error                 => 2000,
      N_Raise_Program_Error                    => 2017,
      N_Raise_Storage_Error                    => 2034,
      N_Integer_Literal                        => 2051,
      N_Real_Literal                           => 2087,
      N_String_Literal                         => 2157,
      N_Explicit_Dereference                   => 2227,
      N_Expression_With_Actions                => 2304,
      N_If_Expression                          => 2323,
      N_Indexed_Component                      => 2388,
      N_Null                                   => 2449,
      N_Qualified_Expression                   => 2449,
      N_Quantified_Expression                  => 2504,
      N_Aggregate                              => 2578,
      N_Allocator                              => 2757,
      N_Case_Expression                        => 2943,
      N_Delta_Aggregate                        => 2985,
      N_Extension_Aggregate                    => 3019,
      N_Raise_Expression                       => 3141,
      N_Range                                  => 3181,
      N_Reference                              => 3222,
      N_Selected_Component                     => 3229,
      N_Slice                                  => 3351,
      N_Target_Name                            => 3373,
      N_Type_Conversion                        => 3373,
      N_Unchecked_Expression                   => 3510,
      N_Unchecked_Type_Conversion              => 3521,
      N_Subtype_Indication                     => 3576,
      N_Component_Declaration                  => 3616,
      N_Entry_Declaration                      => 3686,
      N_Expression_Function                    => 3810,
      N_Formal_Object_Declaration              => 3854,
      N_Formal_Type_Declaration                => 3988,
      N_Full_Type_Declaration                  => 4089,
      N_Incomplete_Type_Declaration            => 4193,
      N_Iterator_Specification                 => 4300,
      N_Loop_Parameter_Specification           => 4371,
      N_Object_Declaration                     => 4435,
      N_Protected_Type_Declaration             => 4730,
      N_Private_Extension_Declaration          => 4833,
      N_Private_Type_Declaration               => 5022,
      N_Subtype_Declaration                    => 5148,
      N_Function_Specification                 => 5269,
      N_Procedure_Specification                => 5401,
      N_Access_Function_Definition             => 5520,
      N_Access_Procedure_Definition            => 5637,
      N_Task_Type_Declaration                  => 5703,
      N_Package_Body_Stub                      => 5801,
      N_Protected_Body_Stub                    => 5880,
      N_Subprogram_Body_Stub                   => 5959,
      N_Task_Body_Stub                         => 6032,
      N_Function_Instantiation                 => 6111,
      N_Procedure_Instantiation                => 6348,
      N_Package_Instantiation                  => 6585,
      N_Package_Body                           => 6790,
      N_Subprogram_Body                        => 6888,
      N_Protected_Body                         => 7221,
      N_Task_Body                              => 7303,
      N_Implicit_Label_Declaration             => 7441,
      N_Package_Declaration                    => 7477,
      N_Single_Task_Declaration                => 7546,
      N_Subprogram_Declaration                 => 7597,
      N_Use_Package_Clause                     => 7706,
      N_Generic_Package_Declaration            => 7822,
      N_Generic_Subprogram_Declaration         => 7919,
      N_Constrained_Array_Definition           => 7992,
      N_Unconstrained_Array_Definition         => 8042,
      N_Exception_Renaming_Declaration         => 8077,
      N_Object_Renaming_Declaration            => 8102,
      N_Package_Renaming_Declaration           => 8215,
      N_Subprogram_Renaming_Declaration        => 8251,
      N_Generic_Function_Renaming_Declaration  => 8340,
      N_Generic_Package_Renaming_Declaration   => 8376,
      N_Generic_Procedure_Renaming_Declaration => 8412,
      N_Abort_Statement                        => 8448,
      N_Accept_Statement                       => 8454,
      N_Assignment_Statement                   => 8549,
      N_Asynchronous_Select                    => 8798,
      N_Block_Statement                        => 8836,
      N_Case_Statement                         => 9095,
      N_Code_Statement                         => 9156,
      N_Compound_Statement                     => 9167,
      N_Conditional_Entry_Call                 => 9175,
      N_Delay_Relative_Statement               => 9214,
      N_Delay_Until_Statement                  => 9225,
      N_Entry_Call_Statement                   => 9236,
      N_Free_Statement                         => 9367,
      N_Goto_Statement                         => 9435,
      N_Loop_Statement                         => 9455,
      N_Null_Statement                         => 9602,
      N_Raise_Statement                        => 9602,
      N_Requeue_Statement                      => 9630,
      N_Simple_Return_Statement                => 9733,
      N_Extended_Return_Statement              => 9856,
      N_Selective_Accept                       => 9985,
      N_Timed_Entry_Call                       => 10021,
      N_Exit_Statement                         => 10062,
      N_If_Statement                           => 10097,
      N_Accept_Alternative                     => 10188,
      N_Delay_Alternative                      => 10264,
      N_Elsif_Part                             => 10316,
      N_Entry_Body_Formal_Part                 => 10360,
      N_Iteration_Scheme                       => 10421,
      N_Terminate_Alternative                  => 10501,
      N_Formal_Abstract_Subprogram_Declaration => 10540,
      N_Formal_Concrete_Subprogram_Declaration => 10579,
      N_Push_Constraint_Error_Label            => 10618,
      N_Push_Program_Error_Label               => 10634,
      N_Push_Storage_Error_Label               => 10650,
      N_Pop_Constraint_Error_Label             => 10666,
      N_Pop_Program_Error_Label                => 10666,
      N_Pop_Storage_Error_Label                => 10666,
      N_SCIL_Dispatch_Table_Tag_Init           => 10666,
      N_SCIL_Dispatching_Call                  => 10678,
      N_SCIL_Membership_Test                   => 10728,
      N_Abortable_Part                         => 10755,
      N_Abstract_Subprogram_Declaration        => 10766,
      N_Access_Definition                      => 10780,
      N_Access_To_Object_Definition            => 10877,
      N_Aspect_Specification                   => 10971,
      N_Call_Marker                            => 11147,
      N_Case_Expression_Alternative            => 11323,
      N_Case_Statement_Alternative             => 11373,
      N_Compilation_Unit                       => 11415,
      N_Compilation_Unit_Aux                   => 11624,
      N_Component_Association                  => 11695,
      N_Component_Definition                   => 11762,
      N_Component_List                         => 11838,
      N_Contract                               => 11880,
      N_Derived_Type_Definition                => 11957,
      N_Decimal_Fixed_Point_Definition         => 12139,
      N_Defining_Program_Unit_Name             => 12199,
      N_Delta_Constraint                       => 12224,
      N_Designator                             => 12258,
      N_Digits_Constraint                      => 12274,
      N_Discriminant_Association               => 12309,
      N_Discriminant_Specification             => 12335,
      N_Enumeration_Type_Definition            => 12425,
      N_Entry_Body                             => 12444,
      N_Entry_Call_Alternative                 => 12551,
      N_Entry_Index_Specification              => 12598,
      N_Exception_Declaration                  => 12646,
      N_Exception_Handler                      => 12714,
      N_Floating_Point_Definition              => 12834,
      N_Formal_Decimal_Fixed_Point_Definition  => 12877,
      N_Formal_Derived_Type_Definition         => 12877,
      N_Formal_Discrete_Type_Definition        => 12975,
      N_Formal_Floating_Point_Definition       => 12975,
      N_Formal_Modular_Type_Definition         => 12975,
      N_Formal_Ordinary_Fixed_Point_Definition => 12975,
      N_Formal_Package_Declaration             => 12975,
      N_Formal_Private_Type_Definition         => 13071,
      N_Formal_Incomplete_Type_Definition      => 13142,
      N_Formal_Signed_Integer_Type_Definition  => 13157,
      N_Freeze_Entity                          => 13157,
      N_Freeze_Generic_Entity                  => 13225,
      N_Generic_Association                    => 13232,
      N_Handled_Sequence_Of_Statements         => 13292,
      N_Index_Or_Discriminant_Constraint       => 13365,
      N_Iterated_Component_Association         => 13377,
      N_Itype_Reference                        => 13450,
      N_Label                                  => 13456,
      N_Modular_Type_Definition                => 13482,
      N_Number_Declaration                     => 13493,
      N_Ordinary_Fixed_Point_Definition        => 13542,
      N_Others_Choice                          => 13584,
      N_Package_Specification                  => 13619,
      N_Parameter_Association                  => 13728,
      N_Parameter_Specification                => 13810,
      N_Pragma                                 => 13978,
      N_Protected_Definition                   => 14347,
      N_Range_Constraint                       => 14399,
      N_Real_Range_Specification               => 14416,
      N_Record_Definition                      => 14437,
      N_Signed_Integer_Type_Definition         => 14608,
      N_Single_Protected_Declaration           => 14629,
      N_Subunit                                => 14685,
      N_Task_Definition                        => 14721,
      N_Triggering_Alternative                 => 14826,
      N_Use_Type_Clause                        => 14873,
      N_Validate_Unchecked_Conversion          => 15009,
      N_Variable_Reference_Marker              => 15033,
      N_Variant                                => 15141,
      N_Variant_Part                           => 15234,
      N_With_Clause                            => 15248,
      N_Unused_At_End                          => 15553);

end Treeprs;
